use std::sync::Arc;

use safer_ffi::derive_ReprC;
use spdlog::{debug, error, info, trace, warn};

#[derive_ReprC]
#[repr(opaque)]
pub struct TsdbEntry {
    table_name: String,
    metric_name: String,
    async_buffer_size: u64,
    inserter_duration: std::time::Duration,
    inserter: Option<clickhouse::inserter::Inserter<MyRowFull<'static>>>,
    client: clickhouse::Client,
    runtime: Arc<tokio::runtime::Runtime>,
}

#[derive_ReprC]
#[repr(opaque)]
pub struct TsdbEntryInsertBatch {
    metric_name: String,
    insert: Option<clickhouse::insert::Insert<MyRowFull<'static>>>,
    runtime: Arc<tokio::runtime::Runtime>,
}

macro_rules! logtry {
    ($e:expr) => {
        match $e {
            Ok(v) => v,
            Err(e) => {
                error!("{}: {}", stringify!($e), e);
                return false;
            }
        }
    };
}

macro_rules! logtry_opt {
    ($e:expr) => {
        match $e {
            Ok(v) => v,
            Err(e) => {
                error!("{}: {}", stringify!($e), e);
                return None;
            }
        }
    };
}

macro_rules! logtrybool {
    ($e:expr) => {
        match $e {
            Ok(_) => true,
            Err(e) => {
                error!("{}: {}", stringify!($e), e);
                false
            }
        }
    };
}

#[derive_ReprC]
#[derive(Debug)]
#[repr(C)]
pub struct FfiMyRow<'a> {
    pub name: safer_ffi::string::str_ref<'a>,
    pub value: f64,
    pub nanoseconds: i64,
}

#[derive(Debug)]
pub struct MyRow<'a> {
    pub name: &'a str,
    pub value: f64,
    pub nanoseconds: i64,
}

#[derive(Debug, clickhouse::Row, serde::Serialize)]
pub struct MyRowFull<'a> {
    pub point_name: &'a str,
    pub metric_name: &'a str,
    pub value: f64,
    pub nanoseconds: i64,
}

impl<'a> MyRowFull<'a> {
    pub fn new(point_name: &'a str, metric_name: &'a str, value: f64, nanoseconds: i64) -> Self {
        Self {
            point_name,
            metric_name,
            value,
            nanoseconds,
        }
    }
}

impl TsdbEntry {
    pub fn new(host: &str, mut port: u16) -> Option<Self> {
        debug!("Connecting to stck at {}:{}", host, port);

        let runtime = logtry_opt!(tokio::runtime::Builder::new_current_thread()
            // .worker_threads(4)
            .enable_all()
            .build());
        let client = runtime.block_on(async {
            if port == 9000 {
                // HACK: TCP to HTTP port
                let no_fix_tcp_port = std::env::var("TSDB_NO_FIX_TCP_PORT")
                    .map(|x| x != "0")
                    .unwrap_or(false);
                if no_fix_tcp_port {
                    debug!("9000 is for TCP port, not fixing");
                } else {
                    debug!("9000 is for TCP port, assuming 8123");
                    port = 8123;
                }
            }
            let url = format!("http://{}:{}", host, port);
            let client = clickhouse::Client::default()
                .with_url(url)
                .with_compression(clickhouse::Compression::Lz4);

            debug!("Connected to stck at {}:{}", host, port);

            client
        });

        let table_name = "tsdb_cpp_dist".to_owned();
        let mut this = Self {
            table_name,
            metric_name: String::new(),
            async_buffer_size: 1,
            inserter_duration: std::time::Duration::from_secs(3),
            inserter: None,
            client,
            runtime: Arc::new(runtime),
        };
        this.inserter = this.create_inserter();

        Some(this)
    }

    pub fn set_active_table(&mut self, table: &str) -> bool {
        debug!("Setting active table to {}", table);
        if self.inserter.is_some() {
            self.commit_insert();
        }
        self.table_name = table.to_owned();
        self.inserter = self.create_inserter();
        self.inserter.is_some()
    }

    pub fn set_metric_name(&mut self, name: &str) -> bool {
        debug!("Setting metric name to {}", name);
        self.metric_name = name.to_owned();
        true
    }

    pub fn set_async_insert_buffer_size(&mut self, size: u64) -> bool {
        debug!("Setting async insert buffer size to {}", size);
        self.async_buffer_size = size;
        if let Some(inserter) = self.inserter.take() {
            self.inserter = Some(inserter.with_max_rows(size));
        }
        true
    }

    pub fn get_async_insert_buffer_size(&self) -> u64 {
        self.async_buffer_size
    }

    pub fn insert_point(&mut self, name: &str, value: f64, nanoseconds: i64) -> bool {
        (self.runtime.clone()).block_on(self.insert_point_async(name, value, nanoseconds))
    }

    pub fn insert_points(&self, points: &[MyRow<'_>]) -> bool {
        self.runtime.block_on(self.insert_points_async(points))
    }

    pub fn insert_points_with_callback(&self, f: impl FnMut() -> FfiMyRow<'static>) -> bool {
        self.runtime
            .block_on(self.insert_points_async_with_callback(f))
    }

    pub fn commit_insert(&mut self) -> bool {
        if let Some(inserter) = &mut self.inserter {
            let stats = logtry!(self.runtime.block_on(inserter.commit()));
            if stats.rows > 0 {
                debug!(
                    "{} bytes, {} rows, {} transactions have been inserted",
                    stats.bytes, stats.rows, stats.transactions
                );
            }
        }
        true
    }

    pub fn force_commit_insert(&mut self) -> bool {
        if let Some(inserter) = &mut self.inserter {
            let stats = logtry!(self.runtime.block_on(inserter.force_commit()));
            if stats.rows > 0 {
                debug!(
                    "{} bytes, {} rows, {} transactions have been inserted",
                    stats.bytes, stats.rows, stats.transactions
                );
            }
        }
        true
    }

    pub fn end_commit_insert(&mut self) -> bool {
        if let Some(inserter) = self.inserter.take() {
            let stats = logtry!(self.runtime.block_on(inserter.end()));
            if stats.rows > 0 {
                debug!(
                    "{} bytes, {} rows, {} transactions have been inserted",
                    stats.bytes, stats.rows, stats.transactions
                );
            }
        } else {
            warn!("No active inserter to commit");
        }
        true
    }

    pub fn new_insert_batch(&self) -> Option<TsdbEntryInsertBatch> {
        self.runtime.block_on(async {
            debug!("Creating new insert batch");
            let insert = logtry_opt!(self.client.insert(&self.table_name));
            Some(TsdbEntryInsertBatch {
                metric_name: self.metric_name.clone(),
                insert: Some(insert),
                runtime: self.runtime.clone(),
            })
        })
    }

    pub async fn insert_point_async(&mut self, name: &str, value: f64, nanoseconds: i64) -> bool {
        debug!(
            "Inserting point {} = {} with t={}ns",
            name, value, nanoseconds
        );

        // let point = MyRowFull::new(name, &self.metric_name, value, nanoseconds);
        // let mut insert = logtry!(self.client.insert(&self.table_name));
        // logtry!(insert.write(&point).await);
        // logtry!(insert.end().await);

        // Use inserter
        let inserter = match &mut self.inserter {
            Some(inserter) => inserter,
            None => {
                error!("No active inserter to insert point");
                return false;
            }
        };
        let point = MyRowFull::new(name, &self.metric_name, value, nanoseconds);
        let point: MyRowFull<'static> = unsafe { std::mem::transmute(point) };
        if let Err(e) = inserter.write(&point) {
            error!("Failed to insert point: {}", e);
            self.inserter = self.create_inserter();
            return false;
        }
        if !self.commit_insert_async().await {
            error!("Failed to commit insert");
            return false;
        }

        debug!(
            "Inserted point {} = {} with t={}ns",
            name, value, nanoseconds
        );
        true
    }

    pub async fn insert_points_async(&self, points: &[MyRow<'_>]) -> bool {
        debug!("Inserting {} points", points.len());
        let mut insert = logtry!(self.client.insert(&self.table_name));
        for point in points {
            let point = MyRowFull::new(
                point.name,
                &self.metric_name,
                point.value,
                point.nanoseconds,
            );
            logtry!(insert.write(&point).await);
        }
        logtry!(insert.end().await);
        debug!("Inserted {} points", points.len());
        true
    }

    pub async fn insert_points_async_with_callback(
        &self,
        mut f: impl FnMut() -> FfiMyRow<'static>,
    ) -> bool {
        debug!("Inserting many points");
        let mut insert = logtry!(self.client.insert(&self.table_name));
        let mut count: u64 = 0;
        loop {
            count += 1;

            let point = f();
            if point.name.is_empty() {
                count -= 1;
                break;
            }
            let point = MyRowFull::new(
                point.name.as_str(),
                &self.metric_name,
                point.value,
                point.nanoseconds,
            );
            logtry!(insert.write(&point).await);
        }
        logtry!(insert.end().await);
        debug!("Inserted many points, total {}", count);
        true
    }

    pub async fn commit_insert_async(&mut self) -> bool {
        if let Some(inserter) = &mut self.inserter {
            let stats = match inserter.commit().await {
                Ok(v) => v,
                Err(e) => {
                    error!("Failed to commit insert: {}", e);
                    self.inserter = self.create_inserter();
                    return false;
                }
            };
            if stats.rows > 0 {
                debug!(
                    "{} bytes, {} rows, {} transactions have been inserted",
                    stats.bytes, stats.rows, stats.transactions
                );
            }
        }
        true
    }

    pub async fn force_commit_insert_async(&mut self) -> bool {
        if let Some(inserter) = &mut self.inserter {
            let stats = match inserter.force_commit().await {
                Ok(v) => v,
                Err(e) => {
                    error!("Failed to force commit insert: {}", e);
                    self.inserter = self.create_inserter();
                    return false;
                }
            };
            if stats.rows > 0 {
                debug!(
                    "{} bytes, {} rows, {} transactions have been inserted",
                    stats.bytes, stats.rows, stats.transactions
                );
            }
        }
        true
    }

    pub async fn end_commit_insert_async(&mut self) -> bool {
        if let Some(inserter) = self.inserter.take() {
            self.inserter = self.create_inserter();
            let stats = logtry!(inserter.end().await);
            if stats.rows > 0 {
                debug!(
                    "{} bytes, {} rows, {} transactions have been inserted",
                    stats.bytes, stats.rows, stats.transactions
                );
            }
        }
        true
    }

    fn create_inserter(&self) -> Option<clickhouse::inserter::Inserter<MyRowFull<'static>>> {
        Some(
            logtry_opt!(self.client.inserter(&self.table_name))
                .with_max_rows(self.async_buffer_size)
                .with_period(Some(self.inserter_duration)),
        )
    }
}

impl TsdbEntryInsertBatch {
    pub fn add_point(&mut self, name: &str, value: f64, nanoseconds: i64) -> bool {
        (self.runtime.clone()).block_on(self.insert_point_async(name, value, nanoseconds))
    }

    pub async fn insert_point_async(&mut self, name: &str, value: f64, nanoseconds: i64) -> bool {
        trace!(
            "Inserting to batch point {} = {} with t={}ns",
            name,
            value,
            nanoseconds
        );

        let point = MyRowFull::new(name, &self.metric_name, value, nanoseconds);
        let point: MyRowFull<'static> = unsafe { std::mem::transmute(point) };
        let Some(mut insert) = self.insert.take() else {
            error!("No active insert batch to insert point");
            return false;
        };
        if let Err(e) = insert.write(&point).await {
            error!("Failed to insert point: {}", e);
            return false;
        }

        trace!(
            "Inserted to batch point {} = {} with t={}ns",
            name,
            value,
            nanoseconds
        );
        self.insert = Some(insert);
        true
    }

    pub fn commit_and_close(&mut self) -> bool {
        if let Some(insert) = self.insert.take() {
            debug!("Committing insert batch");
            if let Err(e) = self.runtime.block_on(insert.end()) {
                error!("Failed to end insert batch: {}", e);
                return false;
            }
            debug!("Committed insert batch");
        }
        true
    }
}

impl Drop for TsdbEntry {
    fn drop(&mut self) {
        if self.inserter.is_some() {
            self.end_commit_insert();
        }
    }
}

impl Drop for TsdbEntryInsertBatch {
    fn drop(&mut self) {
        self.commit_and_close();
    }
}
