use std::panic::AssertUnwindSafe;

use ::safer_ffi::prelude::*;
use safer_ffi::ffi_export;
use spdlog::error;

mod lf;
mod shared;

// NOTE: Used by safer_ffi
#[cfg(feature = "headers")]
pub fn generate_headers() -> ::std::io::Result<()> {
    use ::std::io::Read;
    let builder = ::safer_ffi::headers::builder();
    if let Some(filename) = ::std::env::args_os().nth(1) {
        // Check if the file is unchanged
        let previous_file = if let Ok(file) = ::std::fs::File::open(&filename) {
            let mut reader = ::std::io::BufReader::new(file);
            let mut previous = Vec::new();
            reader.read_to_end(&mut previous)?;
            Some(previous)
        } else {
            None
        };
        let mut new_file = Vec::new();
        builder.to_writer(&mut new_file).generate()?;
        if previous_file.as_deref() == Some(&new_file[..]) {
            // File is unchanged, don't touch it to prevent unnecessary rebuilds
            return Ok(());
        }
        ::std::fs::write(&filename, new_file)
    } else {
        builder.to_writer(::std::io::stdout()).generate()
    }
}

#[ffi_export]
fn rust_log_error_not_open(name: char_p::Ref<'_>) {
    error!("{} is not open", name.to_str());
}

#[ffi_export]
fn rust_log_error_custom(msg: char_p::Ref<'_>) {
    error!("{}", msg.to_str());
}

macro_rules! logpanic {
    ($($arg:tt)*) => {
        // if let Err(e) = std::panic::catch_unwind(|| {
        //     $($arg)*
        // }) {
        //     error!("Rust panic: {:?}", e);
        // }
        match std::panic::catch_unwind(AssertUnwindSafe(move || {
            $($arg)*
        })) {
            Ok(v) => Some(v),
            Err(e) => {
                error!("Rust panic: {:?}", e);
                None
            }
        }
    };
}

#[ffi_export]
fn rust_create_tsdb_lf_entry(
    host: char_p::Ref<'_>,
    port: u16,
) -> Option<repr_c::Box<lf::TsdbEntry>> {
    logpanic!({
        shared::init_tsdb_log();

        lf::TsdbEntry::new(host.to_str(), port)
    })
    .flatten()
    .map(|x| Box::new(x).into())
}

#[ffi_export]
fn rust_drop_tsdb_lf_entry(tsdb: repr_c::Box<lf::TsdbEntry>) {
    drop(tsdb);
}

#[ffi_export]
fn rust_tsdb_lf_entry_set_active_table(tsdb: &mut lf::TsdbEntry, table: char_p::Ref<'_>) -> bool {
    logpanic!(tsdb.set_active_table(table.to_str())).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_set_metric_name(tsdb: &mut lf::TsdbEntry, name: char_p::Ref<'_>) -> bool {
    logpanic!(tsdb.set_metric_name(name.to_str())).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_set_async_insert_buffer_size(tsdb: &mut lf::TsdbEntry, size: u64) -> bool {
    logpanic!(tsdb.set_async_insert_buffer_size(size)).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_get_async_insert_buffer_size(tsdb: &lf::TsdbEntry) -> u64 {
    logpanic!(tsdb.get_async_insert_buffer_size()).unwrap_or(0)
}

#[ffi_export]
fn rust_tsdb_lf_entry_insert_point(
    tsdb: &mut lf::TsdbEntry,
    name: char_p::Ref<'_>,
    value: f64,
    nanoseconds: i64,
) -> bool {
    logpanic!(tsdb.insert_point(name.to_str(), value, nanoseconds)).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_insert_points_with_callback(
    tsdb: &mut lf::TsdbEntry,
    f: extern "C" fn(*mut std::ffi::c_void) -> lf::FfiMyRow<'static>,
    opaque: *mut std::ffi::c_void,
) -> bool {
    logpanic!(tsdb.insert_points_with_callback(|| f(opaque))).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_commit_insert(tsdb: &mut lf::TsdbEntry) -> bool {
    logpanic!(tsdb.commit_insert()).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_force_commit_insert(tsdb: &mut lf::TsdbEntry) -> bool {
    logpanic!(tsdb.force_commit_insert()).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_end_commit_insert(tsdb: &mut lf::TsdbEntry) -> bool {
    logpanic!(tsdb.end_commit_insert()).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_batch_add_point(
    batch: &mut lf::TsdbEntryInsertBatch,
    name: char_p::Ref<'_>,
    value: f64,
    nanoseconds: i64,
) -> bool {
    logpanic!(batch.add_point(name.to_str(), value, nanoseconds)).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_batch_commit_and_close(batch: &mut lf::TsdbEntryInsertBatch) -> bool {
    logpanic!(batch.commit_and_close()).unwrap_or(false)
}

#[ffi_export]
fn rust_tsdb_lf_entry_new_insert_batch(
    tsdb: &lf::TsdbEntry,
) -> Option<repr_c::Box<lf::TsdbEntryInsertBatch>> {
    logpanic!(tsdb.new_insert_batch())
        .flatten()
        .map(|x| Box::new(x).into())
}

#[ffi_export]
fn rust_tsdb_lf_entry_insert_batch_drop(batch: repr_c::Box<lf::TsdbEntryInsertBatch>) {
    drop(batch);
}

// #[cfg(test)]
// mod tests {
//     use super::*;
// }
