use spdlog::sink::RotationPolicy;
use spdlog::sink::StdStream;
use spdlog::sink::{RotatingFileSink, StdStreamSink};
use spdlog::warn;
use spdlog::Level;
use spdlog::LevelFilter;
use spdlog::Logger;
use std::env;
use std::sync::Arc;
use std::sync::OnceLock;

static TSDBLOG_INIT: OnceLock<()> = OnceLock::new();

pub fn init_tsdb_log() {
    TSDBLOG_INIT.get_or_init(|| {
        let log_level = env::var("TSDBLOG_LEVEL").unwrap_or_else(|_| "info".to_string());
        let use_file_logging =
            env::var("TSDB_USE_FILE_LOGGING").unwrap_or_else(|_| "1".to_string()) != "0";
        let log_path = env::var("TSDB_LOG_PATH").unwrap_or_else(|_| "tsdb.log".to_string());

        // init spdlog
        let log_level = match log_level.as_str() {
            "trace" => Level::Trace,
            "debug" => Level::Debug,
            "info" => Level::Info,
            "warn" => Level::Warn,
            "error" => Level::Error,
            "critical" => Level::Critical,
            _ => Level::Info,
        };

        let console_sink = match StdStreamSink::builder()
            .level_filter(spdlog::LevelFilter::MoreSevereEqual(log_level))
            .std_stream(StdStream::Stderr)
            .build()
        {
            Ok(s) => Some(Arc::new(s)),
            Err(e) => {
                eprintln!("[tsdb] Failed to create console sink: {}", e);
                None
            }
        };
        let file_sink = match RotatingFileSink::builder()
            .base_path(&log_path)
            .rotation_policy(RotationPolicy::FileSize(1048576 * 5))
            .max_files(5)
            .level_filter(spdlog::LevelFilter::MoreSevereEqual(log_level))
            .build()
        {
            Ok(s) => Some(Arc::new(s)),
            Err(e) => {
                eprintln!("[tsdb] Failed to create file sink: {}", e);
                None
            }
        };
        let err_file_sink = match RotatingFileSink::builder()
            .base_path(format!("{}.error.log", &log_path))
            .rotation_policy(RotationPolicy::FileSize(1048576 * 5))
            .max_files(5)
            .level_filter(spdlog::LevelFilter::MoreSevereEqual(Level::Error))
            .build()
        {
            Ok(s) => Some(Arc::new(s)),
            Err(e) => {
                eprintln!("[tsdb] Failed to create error file sink: {}", e);
                None
            }
        };

        let mut logger = Logger::builder();
        logger.name("tsdb");
        logger.level_filter(LevelFilter::MoreSevereEqual(log_level));
        if use_file_logging {
            if let Some(console_sink) = console_sink {
                logger.sink(console_sink);
            }
            if let Some(file_sink) = file_sink {
                logger.sink(file_sink);
            }
            if let Some(err_file_sink) = err_file_sink {
                logger.sink(err_file_sink);
            }
        } else {
            if let Some(console_sink) = console_sink {
                logger.sink(console_sink);
            }
            if let Some(err_file_sink) = err_file_sink {
                logger.sink(err_file_sink);
            }
        }

        let logger = match logger.build() {
            Ok(l) => l,
            Err(e) => {
                eprintln!("[tsdb] Failed to create logger: {}", e);
                return;
            }
        };
        spdlog::set_default_logger(Arc::new(logger));

        warn!(
            "tsdb log initialized, compile datetime: {}, rustc version: {}",
            compile_time::datetime_str!(),
            compile_time::rustc_version_str!()
        );
    });
}
